/* 
 * Copyright (c) 2011, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

/*
 *  ======== lpm_driver.c ========
 *
 */


/*  ---------- Operating System headers */

#include <windows.h>
#include <winbase.h>
#include <ceddk.h>

#define __D(c, fmt, ...) DEBUGMSG((c), (L"lpmdrv: " L##fmt, __VA_ARGS__))
#define __E(fmt, ...) RETAILMSG(TRUE, (L"Error: lpmdrv: " L##fmt, __VA_ARGS__))

#define LPM_REG_KEY TEXT("Drivers\\BuiltIn\\LPM")
#define LPM_REG_KEY_TRACE TEXT("Trace")


/*  ---------- LPM header and OMAP3530 macros */

#include <lpm_base.h>

/* OMAP3530 module and mmr addresses (physical) */
#define SYSTEM_CONTROL_ADDR     0x48002000
#define SYSTEM_CONTROL_SIZE     0xC00
#define CONTROL_IVA2_BOOTADDR   0x00000400
#define CONTROL_IVA2_BOOTMOD    0x00000404

#define PRCM_IVA2_CM_ADDR       0x48004000
#define PRCM_IVA2_CM_SIZE       0x800
#define CM_FCLKEN_IVA2          0x00000000
#define CM_CLKEN_PLL_IVA2       0x00000004
#define CM_IDLEST_IVA2          0x00000020
#define CM_AUTOIDLE_PLL_IVA2    0x00000034
#define CM_CLKSEL1_PLL_IVA2     0x00000040
#define CM_CLKSEL2_PLL_IVA2     0x00000044
#define CM_CLKSTCTRL_IVA2       0x00000048

#define PRCM_IVA2_PRM_ADDR      0x48306000
#define PRCM_IVA2_PRM_SIZE      0x800
#define RM_RSTCTRL_IVA2         0x00000050
#define RM_RSTST_IVA2           0x00000058
#define PM_WKDEP_IVA2           0x000000C8
#define PM_PWSTCTRL_IVA2        0x000000E0
#define PM_PWSTST_IVA2          0x000000E4

#define REG(x) *((volatile unsigned int *) (x))
#define MAX_WAIT_COUNT 0x50000


typedef struct deviceState {
    BOOL        initialized;    /* device state is valid */
    ULONG       openCount;      /* number of open clients */
    PVOID       controlVA;      /* system control module */
    PVOID       ivacmVA;        /* iva clock manager */
    PVOID       ivaprmVA;       /* iva power and reset manager */
    INT         gblOnRefCount;  /* global reference 'power-on' count */
} DeviceState;

DeviceState LpmDevice;

typedef struct clientState {
    DeviceState *devState;      /* pointer to device state */
    UINT        onRefCount;     /* client reference count */
    UINT        state;          /* client state mask */
} ClientState;


int LpmDevOff(ClientState *pLpmInst);
int LpmDevOn(ClientState *pLpmInst);
void LpmGetTraceValue(void);

#if defined(DEBUG)
DWORD dwT = 0;  /* global trace flag */
#endif


/*
 *  ======== LPM_Init ========
 *  Called at system boot time when driver is loaded.
 */

DWORD LPM_Init(PVOID context)
{
    PHYSICAL_ADDRESS pa;
    DeviceState *pLpmDev = &LpmDevice;

#if defined(DEBUG)
    /* get LPM/Trace value from registry */
    LpmGetTraceValue();
#endif

    __D(dwT, "--> LPM_Init\n");

    /* initialize driver state */
    pLpmDev->initialized = FALSE;
    pLpmDev->openCount = 0;
    pLpmDev->controlVA = NULL;
    pLpmDev->ivacmVA = NULL;
    pLpmDev->ivaprmVA = NULL;
    pLpmDev->gblOnRefCount = 0;

    /* map hardware modules into driver's address space */
    pa.QuadPart = SYSTEM_CONTROL_ADDR;
    pLpmDev->controlVA = MmMapIoSpace(pa, SYSTEM_CONTROL_SIZE, FALSE);
    if (NULL == pLpmDev->controlVA) {
        __E("failed to map System Control Module\n");
        goto fail;
    }

    pa.QuadPart = PRCM_IVA2_CM_ADDR;
    pLpmDev->ivacmVA = MmMapIoSpace(pa, PRCM_IVA2_CM_SIZE, FALSE);
    if (NULL == pLpmDev->ivacmVA) {
        __E("failed to map IVA Clock Manager\n");
        goto fail;
    }

    pa.QuadPart = PRCM_IVA2_PRM_ADDR;
    pLpmDev->ivaprmVA = MmMapIoSpace(pa, PRCM_IVA2_PRM_SIZE, FALSE);
    if (NULL == pLpmDev->ivaprmVA) {
        __E("failed to map IVA Power & Reset Manager\n");
        goto fail;
    }

    __D(dwT, "System Control Module: 0x%X (pa) -> 0x%X (va)\n",
        SYSTEM_CONTROL_ADDR, pLpmDev->controlVA);
    __D(dwT, "IVA2 Clock Manager: 0x%X (pa) -> 0x%X (va)\n",
        PRCM_IVA2_CM_ADDR, pLpmDev->ivacmVA);
    __D(dwT, "IVA2 Power & Reset Manager: 0x%X (pa) -> 0x%X (va)\n",
        PRCM_IVA2_PRM_ADDR, pLpmDev->ivaprmVA);

    /* driver successfully initialized */
    pLpmDev->initialized = TRUE;

    __D(dwT, "<-- LPM_Init\n");
    return((DWORD)pLpmDev);

fail:
    __D(dwT, "<-- LPM_Init (failure state)\n");
    return(0);
}


/*
 *  ======== LPM_Deinit ========
 */
BOOL LPM_Deinit(DWORD dwContext)
{
    DeviceState *pLpmDev = (DeviceState *)dwContext;

#if defined(DEBUG)
    /* get LPM/Trace value from registry */
    LpmGetTraceValue();
#endif

    __D(dwT, "--> LPM_Deinit\n");

    /* unmap virtual address windows */
    MmUnmapIoSpace((PVOID)SYSTEM_CONTROL_ADDR, SYSTEM_CONTROL_SIZE);
    pLpmDev->controlVA = NULL;
    MmUnmapIoSpace((PVOID)PRCM_IVA2_CM_ADDR, PRCM_IVA2_CM_SIZE);
    pLpmDev->ivacmVA = NULL;
    MmUnmapIoSpace((PVOID)PRCM_IVA2_PRM_ADDR, PRCM_IVA2_PRM_SIZE);
    pLpmDev->ivaprmVA = NULL;

    /* check for open clients (should not have any) */
    if (0 != pLpmDev->openCount) {
        __E("driver unloaded with open clients\n");
    }

    /* driver has been de-initialized */
    pLpmDev->initialized = FALSE;
    __D(dwT, "<-- LPM_Deinit\n");
    return(TRUE);
}


/*
 *  ======== LPM_Open ========
 */
DWORD LPM_Open(DWORD deviceContext, DWORD access, DWORD shareMode)
{
    DeviceState *pLpmDev = (DeviceState *)deviceContext;
    ClientState *pLpmInst = NULL;

#if defined(DEBUG)
    /* get LPM/Trace value from registry */
    LpmGetTraceValue();
#endif

    __D(dwT, "--> LPM_Open\n");

    /* create a new client instance */
    pLpmInst = (ClientState *)LocalAlloc(LPTR, sizeof(ClientState));
    if (NULL == pLpmInst) {
        __E("failed to allocate client object\n");
        goto fail;
    }

    /* initialize client state */
    pLpmInst->devState = pLpmDev;
    pLpmInst->onRefCount = 0;
    pLpmInst->state = 0;

    /* increment the open client count */
    InterlockedIncrement(&pLpmDev->openCount);
    __D(dwT, "client open count = %d\n", pLpmDev->openCount);

    __D(dwT, "<-- LPM_Open\n");
    return((DWORD)pLpmInst);

fail:
    __D(dwT, "<-- LPM_Open (failure state)\n");
    return(0);
}


/*
 *  ======== LPM_Close ========
 */
BOOL LPM_Close(DWORD clientContext)
{
    ClientState *pLpmInst = (ClientState *)clientContext;
    DeviceState *pLpmDev = pLpmInst->devState;

#if defined(DEBUG)
    /* get LPM/Trace value from registry */
    LpmGetTraceValue();
#endif

    __D(dwT, "--> LPM_Close\n");

    /* verify there is one open client */
    if (1 > pLpmDev->openCount) {
        __E("close failed, no open clients\n");
        goto fail;
    }

    /* decrement the open client count */
    InterlockedDecrement(&pLpmDev->openCount);
    __D(dwT, "client open count = %d\n", pLpmDev->openCount);

    /* free the client instance */
    pLpmInst = LocalFree((HLOCAL)pLpmInst);
    if (NULL != pLpmInst) {
        /* log error but return success */
        __E("failed to free client object\n");
    }

    __D(dwT, "<-- LPM_Close\n");
    return(TRUE);

fail:
    __D(dwT, "<-- LPM_Close (failure state)\n");
    return(FALSE);
}


/*
 *  ======== LPM_IOControl ========
 */
BOOL LPM_IOControl(
    DWORD  dwContext,
    DWORD  ioctl,
    PUCHAR pInBuf,
    DWORD  InBufLen,
    PUCHAR pOutBuf,
    DWORD  OutBufLen,
    PDWORD pdwBytesTransferred)
{
    ClientState *pLpmInst = (ClientState *)dwContext;
    DeviceState *pLpmDev = pLpmInst->devState;
    UINT args;

#if defined(DEBUG)
    /* get LPM/Trace value from registry */
    LpmGetTraceValue();
#endif

    __D(dwT, "--> LPM_IOControl: cmd: 0x%X\n", ioctl);

    /* dispatch requested command */
    switch (ioctl) {
        case LPM_CTRL_CLEAR:
            args = *(LPUINT)pInBuf;
            __D(dwT, "cmd: CLEAR, args: 0x%X\n", args);
            pLpmInst->state &= ~args;
            break;

        case LPM_CTRL_CONNECT:
            __D(dwT, "cmd: CONNECT (not implemented!)\n");
            break;

        case LPM_CTRL_DISCONNECT:
            __D(dwT, "cmd: DISCONNECT (not implemented!)\n");
            break;

        case LPM_CTRL_OFF:
            __D(dwT, "cmd: OFF\n");

            /* ignore reference count if override bit is set */
            if (pLpmInst->state & LPM_CTRL_REFCOUNTOVR) {
                __D(dwT, "using reference count override");
                __D(dwT, "global power-on count = %d\n",pLpmDev->gblOnRefCount);
                __D(dwT, "client power-on count = %d\n", pLpmInst->onRefCount);
                LpmDevOff(pLpmInst);
            }
            else {
                /* decrement the client's 'power-on' reference count */
                pLpmInst->onRefCount--;

                /* decrement the global 'power-on' reference count */
                pLpmDev->gblOnRefCount--;

                __D(dwT, "global power-on count = %d\n",pLpmDev->gblOnRefCount);
                __D(dwT, "client power-on count = %d\n", pLpmInst->onRefCount);

                /* turn off the power if last client */
                if (0 == pLpmDev->gblOnRefCount) {
                    LpmDevOff(pLpmInst);
                }
            }
            break;

        case LPM_CTRL_ON:
            __D(dwT, "cmd: ON\n");

            /* ignore reference count if override bit is set */
            if (pLpmInst->state & LPM_CTRL_REFCOUNTOVR) {
                __D(dwT, "using reference count override");
                __D(dwT, "global power-on count = %d\n",pLpmDev->gblOnRefCount);
                __D(dwT, "client power-on count = %d\n", pLpmInst->onRefCount);
                LpmDevOn(pLpmInst);
            }
            else {
                /* increment the client's 'power-on' reference count */
                pLpmInst->onRefCount++;

                /* increment the global 'power-on' reference count */
                pLpmDev->gblOnRefCount++;

                __D(dwT, "global power-on count = %d\n",pLpmDev->gblOnRefCount);
                __D(dwT, "client power-on count = %d\n", pLpmInst->onRefCount);

                /* turn on the power if first client */
                if (1 == pLpmDev->gblOnRefCount) {
                    LpmDevOn(pLpmInst);
                }
            }
            break;

        case LPM_CTRL_RESUME:
            __D(dwT, "cmd: RESUME (not implemented!)\n");
            break;

        case LPM_CTRL_SETPOWERSTATE:
            __D(dwT, "cmd: SETPOWERSTATE (not implemented!)\n");
            break;

        case LPM_CTRL_SET:
            args = *(LPUINT)pInBuf;
            __D(dwT, "cmd: SET, args: 0x%X\n", args);
            pLpmInst->state |= args;
            break;

        default:
            __E("unknown ioctl received (0x%X)\n", ioctl);
            return FALSE;
    }

    __D(dwT, "<-- LPM_IOControl\n");
    return(TRUE);
}


/*
 *  ======== LpmDevOff ========
 */
int LpmDevOff(ClientState *pLpmInst)
{
    int i;
    int status = LPM_SOK;
    DeviceState *pLpmDev = pLpmInst->devState;
    unsigned int cm = (unsigned int)pLpmDev->ivacmVA;
    unsigned int prm = (unsigned int)pLpmDev->ivaprmVA;
    unsigned int control = (unsigned int)pLpmDev->controlVA;


    __D(dwT, "--> LpmDevOff\n");

    /* MPU controlled ON + ACTIVE */

    // clear power state change request
    REG(cm + CM_CLKSTCTRL_IVA2) = 0;
    REG(cm + CM_CLKSTCTRL_IVA2);

    // turn off IVA2 domain functional clock (required to assert reset)
    REG(cm + CM_FCLKEN_IVA2) = 0;
    REG(cm + CM_FCLKEN_IVA2);

    // assert reset on DSP, SEQ
    REG(prm + RM_RSTCTRL_IVA2) = 0x5;
    REG(prm + RM_RSTCTRL_IVA2);

    // assert reset on MMU, interconnect, etc.
    REG(prm + RM_RSTCTRL_IVA2) |= (1 << 1);
    REG(prm + RM_RSTCTRL_IVA2);

    // set next IVA2 power state to ON
    REG(prm + PM_PWSTCTRL_IVA2) |= (3 << 0);
    REG(prm + PM_PWSTCTRL_IVA2);

    // start a software supervised wake-up transition
    REG(cm + CM_CLKSTCTRL_IVA2) = 0x2;
    REG(cm + CM_CLKSTCTRL_IVA2);

    // wait for IVA2 power domain to switch ON
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(prm + PM_PWSTST_IVA2) & 0x3) == 0x3) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2 power domain not ON in LpmDevOff\n");
        status = LPM_EFAIL;
        goto fail;
    }

    // clear power-on transition request
    REG(cm + CM_CLKSTCTRL_IVA2) = 0;
    REG(cm + CM_CLKSTCTRL_IVA2);

    // turn on IVA2 domain functional clock
    REG(cm + CM_FCLKEN_IVA2) = 1;
    REG(cm + CM_FCLKEN_IVA2);

    // ensure the IVA2_SW_RST2 status bit is cleared
    REG(prm + RM_RSTST_IVA2) |= (1 << 9);
    REG(prm + RM_RSTST_IVA2);

    // release IVA2_RST2 reset (MMU, SYSC, local interconnect, etc)
    REG(prm + RM_RSTCTRL_IVA2) &= ~(1 << 1);
    REG(prm + RM_RSTCTRL_IVA2);

    // wait for IVA2_RST2 signal to be released
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(prm + RM_RSTST_IVA2) & (1 << 9))) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2_RST2 signal was not released\n");
        status = LPM_EFAIL;
        goto fail;
    }

    // clear the IVA2_SW_RST2 status bit
    REG(prm + RM_RSTST_IVA2) |= (1 << 9);
    REG(prm + RM_RSTST_IVA2);

    // wait for IVA2 state to become active (~2 loops @ 312MHz)
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(cm + CM_IDLEST_IVA2) & 0x1) == 0) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2 did not become active\n");
        status = LPM_EFAIL;
        goto fail;
    }


    /* MPU controlled OFF */

    // turn off IVA2 domain functional clock (required to assert reset)
    REG(cm + CM_FCLKEN_IVA2) = 0;
    REG(cm + CM_FCLKEN_IVA2);

    // assert reset on MMU, interconnect, etc.
    REG(prm + RM_RSTCTRL_IVA2) |= (1 << 1);
    REG(prm + RM_RSTCTRL_IVA2);

    // wait for IVA2 state to become idle
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(cm + CM_IDLEST_IVA2) & 0x1) == 0x1) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2 did not become idle\n");
        status = LPM_EFAIL;
        goto fail;
    }

    // set next IVA2 power state to OFF
    REG(prm + PM_PWSTCTRL_IVA2) &= ~(0x3);
    REG(prm + PM_PWSTCTRL_IVA2);

    // start a software supervised sleep transition
    REG(cm + CM_CLKSTCTRL_IVA2) = 0x1;
    REG(cm + CM_CLKSTCTRL_IVA2);

    // wait for IVA2 power domain to switch OFF
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(prm + PM_PWSTST_IVA2) & 0x3) == 0) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2 power domain not OFF (1)\n");
        status = LPM_EFAIL;
        goto fail;
    }

    // clear power state change request
    REG(cm + CM_CLKSTCTRL_IVA2) = 0;
    REG(cm + CM_CLKSTCTRL_IVA2);


    /* MPU controlled ON + ACTIVE */

    // set next IVA2 power state to ON
    REG(prm + PM_PWSTCTRL_IVA2) |= (3 << 0);
    REG(prm + PM_PWSTCTRL_IVA2);

    // start a software supervised wake-up transition
    REG(cm + CM_CLKSTCTRL_IVA2) = 0x2;
    REG(cm + CM_CLKSTCTRL_IVA2);

    // wait for IVA2 power domain to switch ON
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(prm + PM_PWSTST_IVA2) & 0x3) == 0x3) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2 power domain not ON in LPM_on\n");
        status = LPM_EFAIL;
        goto fail;
    }

    // clear power-on transition request
    REG(cm + CM_CLKSTCTRL_IVA2) = 0;
    REG(cm + CM_CLKSTCTRL_IVA2);

    // turn on IVA2 domain functional clock
    REG(cm + CM_FCLKEN_IVA2) = 1;
    REG(cm + CM_FCLKEN_IVA2);

    // ensure the IVA2_SW_RST2 status bit is cleared
    REG(prm + RM_RSTST_IVA2) |= (1 << 9);
    REG(prm + RM_RSTST_IVA2);

    // release IVA2_RST2 reset (MMU, SYSC, local interconnect, etc)
    REG(prm + RM_RSTCTRL_IVA2) &= ~(1 << 1);
    REG(prm + RM_RSTCTRL_IVA2);

    // wait for IVA2_RST2 signal to be released
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(prm + RM_RSTST_IVA2) & (1 << 9))) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2_RST2 signal was not released\n");
        status = LPM_EFAIL;
        goto fail;
    }

    // clear the IVA2_SW_RST2 status bit
    REG(prm + RM_RSTST_IVA2) |= (1 << 9);
    REG(prm + RM_RSTST_IVA2);

    // first wait for IVA2 state to become active
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(cm + CM_IDLEST_IVA2) & 0x1) == 0) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2 did not become active\n");
        status = LPM_EFAIL;
        goto fail;
    }

    // set DSP boot address and boot mode
    REG(control + CONTROL_IVA2_BOOTADDR) = 0;
    REG(control + CONTROL_IVA2_BOOTADDR);
    REG(control + CONTROL_IVA2_BOOTMOD) = 1;
    REG(control + CONTROL_IVA2_BOOTMOD);

    // enable automatic clock gating
    REG(cm + CM_CLKSTCTRL_IVA2) = 3;
    REG(cm + CM_CLKSTCTRL_IVA2);

    // set next IVA2 power state to OFF
    REG(prm + PM_PWSTCTRL_IVA2) &= ~(0x3);
    REG(prm + PM_PWSTCTRL_IVA2);

    // ensure the IVA2_SW_RST1 status bit is cleared
    REG(prm + RM_RSTST_IVA2) |= (1 << 8);
    REG(prm + RM_RSTST_IVA2);

    // release DSP from reset
    REG(prm + RM_RSTCTRL_IVA2) &= ~(1 << 0);
    REG(prm + RM_RSTCTRL_IVA2);

    // wait for IVA2_RST1 signal to be released
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(prm + RM_RSTST_IVA2) & (1 << 8))) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2_RST1 signal was not released\n");
        status = LPM_EFAIL;
        goto fail;
    }

    // clear the IVA2_SW_RST1 status bit
    REG(prm + RM_RSTST_IVA2) |= (1 << 8);
    REG(prm + RM_RSTST_IVA2);

    // wait for IVA2 power domain to switch OFF (~7 loops @ 312MHz)
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(prm + PM_PWSTST_IVA2) & 0x3) == 0) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2 power domain not OFF (2)\n");
        status = LPM_EFAIL;
        goto fail;
    }

    // assert reset on DSP, SEQ
    REG(prm + RM_RSTCTRL_IVA2) = 0x5;
    REG(prm + RM_RSTCTRL_IVA2);

    // assert reset on MMU, interconnect, etc.
    REG(prm + RM_RSTCTRL_IVA2) |= (1 << 1);
    REG(prm + RM_RSTCTRL_IVA2);

    // disable automatic clock gating
    REG(cm + CM_CLKSTCTRL_IVA2) = 0;
    REG(cm + CM_CLKSTCTRL_IVA2);

    // turn off IVA2 domain functional clock
    REG(cm + CM_FCLKEN_IVA2) = 0;
    REG(cm + CM_FCLKEN_IVA2);

    __D(dwT, "<-- LpmDevOff\n");
    return(status);

fail:
    __D(dwT, "<-- LpmDevOff (failure state)\n");
    return(status);
}


/*
 *  ======== LpmDevOn ========
 */
int LpmDevOn(ClientState *pLpmInst)
{
    int i;
    int status = LPM_SOK;
    DeviceState *pLpmDev = pLpmInst->devState;
    unsigned int cm = (unsigned int)pLpmDev->ivacmVA;
    unsigned int prm = (unsigned int)pLpmDev->ivaprmVA;


    __D(dwT, "--> LpmDevOn\n");

    // ensure a clean power-off transition
    status = LpmDevOff(pLpmInst);
    if (LPM_SOK != status) {
        goto fail;
    }

    // disable automatic clock gating
    REG(cm + CM_CLKSTCTRL_IVA2) = 0;
    REG(cm + CM_CLKSTCTRL_IVA2);

    // turn off IVA2 domain functional clock
    REG(cm + CM_FCLKEN_IVA2) = 0;
    REG(cm + CM_FCLKEN_IVA2);

    // assert reset on DSP, SEQ
    REG(prm + RM_RSTCTRL_IVA2) = 0x5;
    REG(prm + RM_RSTCTRL_IVA2);

    // assert reset on MMU, interconnect, etc.
    REG(prm + RM_RSTCTRL_IVA2) |= (1 << 1);
    REG(prm + RM_RSTCTRL_IVA2);

    // set next IVA2 power state to ON
    REG(prm + PM_PWSTCTRL_IVA2) |=
        (3 << 20) |     // SharedL2CacheFlatOnState = ON
        (3 << 0);       // PowerState = ON
    REG(prm + PM_PWSTCTRL_IVA2);

    // start a software supervised wake-up transition
    REG(cm + CM_CLKSTCTRL_IVA2) = 0x2;
    REG(cm + CM_CLKSTCTRL_IVA2);

    // wait for IVA2 power domain to switch ON
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(prm + PM_PWSTST_IVA2) & 0x3) == 0x3) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2 power domain not ON in LpmDevOn\n");
        status = LPM_EFAIL;
        goto fail;
    }

    // clear power-on transition request
    REG(cm + CM_CLKSTCTRL_IVA2) = 0;
    REG(cm + CM_CLKSTCTRL_IVA2);

    // turn on IVA2 domain functional clock
    REG(cm + CM_FCLKEN_IVA2) = 1;
    REG(cm + CM_FCLKEN_IVA2);

    // ensure the IVA2_SW_RST2 status bit is cleared
    REG(prm + RM_RSTST_IVA2) |= (1 << 9);
    REG(prm + RM_RSTST_IVA2);

    // release IVA2_RST2 reset (MMU, SYSC, local interconnect, etc)
    REG(prm + RM_RSTCTRL_IVA2) &= ~(1 << 1);
    REG(prm + RM_RSTCTRL_IVA2);

    // wait for IVA2_RST2 signal to be released
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(prm + RM_RSTST_IVA2) & (1 << 9))) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2_RST2 signal was not released\n");
        status = LPM_EFAIL;
        goto fail;
    }

    // clear the IVA2_SW_RST2 status bit
    REG(prm + RM_RSTST_IVA2) |= (1 << 9);
    REG(prm + RM_RSTST_IVA2);

    // wait for IVA2 state to become active
    for (i = 0; i < MAX_WAIT_COUNT; i++) {
        if ((REG(cm + CM_IDLEST_IVA2) & 0x1) == 0) {
            break;
        }
    }
    if (i == MAX_WAIT_COUNT) {
        __E("IVA2 did not become active\n");
        status = LPM_EFAIL;
        goto fail;
    }

    // enable automatic hardware supervised transitions
    REG(cm + CM_CLKSTCTRL_IVA2) = 0x3;
    REG(cm + CM_CLKSTCTRL_IVA2);

    __D(dwT, "<-- LpmDevOn\n");
    return(status);

fail:
    __D(dwT, "<-- LpmDevOn (failure state)\n");
    return(status);
}


#if defined(DEBUG)
/*
 *  ======== LpmGetTraceValue ========
 */
void LpmGetTraceValue(void)
{
    HKEY hKey;
    DWORD dwSize = sizeof(DWORD);


    /* get LPM/Trace value from registry */
    if (ERROR_SUCCESS !=
        RegOpenKeyEx(HKEY_LOCAL_MACHINE, LPM_REG_KEY, 0, 0, &hKey)) {

        __E("cannot open registry key HKEY_LOCAL_MACHINE\\%s\n", LPM_REG_KEY);
        dwT = 0;
        goto fail;
    }

    if (ERROR_SUCCESS !=
        RegQueryValueEx(hKey, LPM_REG_KEY_TRACE, NULL, NULL,
            (LPBYTE)&dwT, &dwSize)) {

        /* log error and continue */
        __E("cannot query registry value HKEY_LOCAL_MACHINE\\%s\\%s\n",
            LPM_REG_KEY, LPM_REG_KEY_TRACE);
        dwT = 0;
    }

    RegCloseKey(hKey);

fail:
    return;
}
#endif
/*
 *  @(#) ti.bios.power; 1, 1, 1,1; 12-9-2011 14:13:01; /db/atree/library/trees/power/power-g11x/src/ xlibrary

 */

